/* FLENG - Linux specific part of the runtime system */

#include "fleng.h"
#include "fleng-util.h"
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <sys/signalfd.h>
#include <sys/timerfd.h>
#include <sys/wait.h>
#include <string.h>
#include <pthread.h>
#include <poll.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>

#define MAX_EVENTS  32

static struct { char *name; int sig; } signals[] = {
  { "SIGHUP", 1 },
  { "SIGINT", 2 },
  { "SIGQUIT", 3 },
  { "SIGILL", 4 },
  { "SIGTRAP", 5 },
  { "SIGABRT", 6 },
  { "SIGIOT", 6 },
  { "SIGFPE", 8 },
  { "SIGKILL", 9 },
  { "SIGSEGV", 11 },
  { "SIGPIPE", 13 },
  { "SIGALRM", 14 },
  { "SIGTERM", 15 },
  { "SIGTTIN", 21 },
  { "SIGTTOU", 22 },
  { "SIGXCPU", 24 },
  { "SIGXFSZ", 25 },
  { "SIGVTALRM", 26 },
  { "SIGPROF", 27 },
  { "SIGWINCH", 28 },
  { "SIGBUS", 7 },
  { "SIGUSR1", 10 },
  { "SIGCHLD", 17 },
  { "SIGCONT", 18 },
  { "SIGIO", 29 },
  { "SIGSTOP", 19 },
  { "SIGSYS", 31 },
  { "SIGTSTP", 20 },
  { "SIGURG", 23 },
  { "SIGUSR2", 12 },
  { "SIGSTKSZ", 0x2000 },
  { "SIGPOLL", 29 },
  { "SIGPWR", 30 },
  { "SIGSTKFLT", 16 },
  { "SIGUNUSED", 31 },
  { "SIGRTMAX", 0 },
  { "SIGRTMIN", 0 },
  { NULL, 0 }
};

struct child {
    pid_t pid;
    FL_TCB *tcb;
    FL_VAL var;
    int running;
    int status;
    struct child *next;
};

static volatile int sigchild = 0, sigint = 0, sigterm = 0;
static volatile int sigint_enabled = 1, sigterm_enabled = 1;
static pthread_mutex_t child_mutex = PTHREAD_MUTEX_INITIALIZER;
static struct child *children;

static void signal_handler(int sig) 
{
    switch(sig) {
    case SIGCHLD: 
        sigchild = 1;
        break;
    case SIGINT:
        sigint = 1;
        break;
    case SIGTERM:
        sigterm = 1;
        break;
    }
}

int fl_create_event_queue(FL_TCB *tcb)
{
    tcb->pollfds = (struct pollfd *)malloc(sizeof(struct pollfd) * MAX_EVENTS);
    struct sigaction sa;
    sa.sa_handler = signal_handler;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;
    sigaction(SIGINT, &sa, NULL);
    sigaction(SIGCHLD, &sa, NULL);
    sigaction(SIGTERM, &sa, NULL);
    sigset_t ss;
    sigfillset(&ss);
    sigdelset(&ss, SIGINT);
    sigdelset(&ss, SIGCHLD);
    sigdelset(&ss, SIGTERM);
    pthread_sigmask(SIG_BLOCK, &ss, NULL);
    children = NULL;
    return 0;
}

static void wake_children(FL_TCB *tcb) 
{
    if(sigchild) {
        sigchild = 0;
        for(;;) {
            int s;
            pid_t w = waitpid(-1, &s, WNOHANG);
            if(w <= 0) break;
            pthread_mutex_lock(&child_mutex);
            for(struct child *c = children; c != NULL; c = c->next) {
                if(c->pid == w) {
                    c->running = 0;
                    if(!WIFEXITED(s)) s = 255; /* XXX ? */
                    else s = WEXITSTATUS(s);
                    c->status = s;
                }
            }
            pthread_mutex_unlock(&child_mutex);
        }
    }
    if(children == NULL) return;
    pthread_mutex_lock(&child_mutex);
    struct child **prev = &children;
    for(struct child *c = children; c != NULL; c = c->next) {
        if(c->tcb == tcb && !c->running) {
            fl_assign(tcb, MKINT(c->status), c->var);
            /* var is kept in listeners, no need to unref */
            *prev = c->next;
            free(c);
        } else prev = &c->next;
    }
    pthread_mutex_unlock(&child_mutex);
}

static int update_events(FL_TCB *tcb, long ms)
{
    int i = 0;
    struct pollfd *pollfds = tcb->pollfds;
    /* XXX re-use pollfds if listeners didn't change */
    for(FL_VAL ll = tcb->listening; ll != fl_nil; ll = CDR(ll)) {
        assert(i < MAX_EVENTS);
        FL_VAL litem = CAR(ll);
        int fd = INT(CAR(litem));
        if(INT(CAR(CDR(CDR(litem)))) != FL_CHILD) {
            pollfds[ i ].fd = fd;
            pollfds[ i ].events = POLLIN;
        }
        ++i;
    }
    int r = poll(pollfds, i, ms);
    if(r == -1 && errno != EINTR) fl_rt_error(tcb, fl_nil, FL_IO_ERROR);
    int intr = errno == EINTR;
    if(sigint && sigint_enabled) exit(2);
    if(sigterm && sigterm_enabled) exit(2);
    wake_children(tcb);
    FL_VAL ll = tcb->listening; 
    FL_VAL *prev = &tcb->listening;
    while(ll != fl_nil) {
        FL_VAL litem = CAR(ll);
        FL_VAL next = CDR(ll);
        int fd = INT(CAR(litem));
        int event = INT(CAR(CDR(CDR(litem))));
        FL_VAL var = CAR(CDR(litem));
        if(event == FL_CHILD) {
            if(CAR(var) != var) { /* bound by wake_children()? */
                *prev = next;
                CDR(ll) = fl_nil;
                unref(tcb, ll);
            } else prev = &CDR(ll);
        } else if(!intr) {
            for(int i = 0; i < r; ++i) {
                if(fd == pollfds[ i ].fd && 
                   (pollfds[ i ].revents & (POLLIN | POLLHUP)) != 0) {
                    uint64_t n = 0;
                    struct signalfd_siginfo si;
                    switch(event) {
                    case FL_TIMEOUT:
                        read(fd, &n, sizeof(uint64_t));
                        if(CAR(var) == var) fl_assign(tcb, fl_nil, var);
                        close(fd);
                        break;
                    case FL_SIGNAL:
                        while(read(fd, &si, sizeof(struct signalfd_siginfo)) > 0) {
                            ++n;
                            if(si.ssi_signo == SIGINT) sigint_enabled = 1;
                            else if(si.ssi_signo == SIGTERM) sigterm_enabled = 1;
                        }
                        if(CAR(var) == var) fl_assign(tcb, MKINT(n), var);
                        close(fd);
                        break;
                    case FL_INPUT:
                        fl_assign(tcb, fl_nil, var);
                        break;
                    }
                    *prev = next;
                    CDR(ll) = fl_nil;
                    unref(tcb, ll);
                    i = r;
                }
            }
        } else prev = &CDR(ll);
        ll = next;
    }
    if(intr) return 0;
    return r;
}

int fl_wait_for_events(FL_TCB *tcb, long ms)
{
    return update_events(tcb, ms);
}

int fl_add_event_to_queue(FL_TCB *tcb, int event, FL_VAL data, FL_VAL var, FL_VAL eid)
{
    /* listener record: [fd, var, event|data] */
    FL_VAL n1 = fl_alloc_cell(tcb, MKINT(event), data);
    ((FL_CELL *)n1)->tag = FL_LIST_TAG;
    n1 = fl_alloc_cell(tcb, var, n1);
    ((FL_CELL *)n1)->tag = FL_LIST_TAG;
    n1 = fl_alloc_cell(tcb, fl_nil, n1);
    ((FL_CELL *)n1)->tag = FL_LIST_TAG;
    FL_VAL n = fl_alloc_cell(tcb, n1, tcb->listening);
    unref(tcb, tcb->listening);
    ((FL_CELL *)n)->tag = FL_LIST_TAG | 1;
    tcb->listening = n;
    sigset_t sigs;
    int fd;
    switch(event) {
    case FL_INPUT:  /* data: fd */
        fd = INT(data);
        break;
    case FL_SIGNAL: { /* data: signal */
        int i;
        for(i = 0; signals[ i ].name != NULL; ++i) {
            if(!strcmp(signals[ i ].name, (char *)data + 1)) break;
        }
        if(signals[ i ].name == NULL) fl_rt_error(tcb, data, FL_BAD_SIGNAL);
        sigemptyset(&sigs);
        sigaddset(&sigs, signals[ i ].sig);
        if(signals[ i ].sig == SIGINT) sigint_enabled = 0;
        if(signals[ i ].sig == SIGTERM) sigterm_enabled = 0;
        fd = signalfd(-1, &sigs, SFD_CLOEXEC | SFD_NONBLOCK);
        break; }
    case FL_TIMEOUT: {   /* data: milliseconds */
        struct itimerspec it;
        fd = timerfd_create(CLOCK_MONOTONIC, TFD_CLOEXEC);
        if(fd < 0) fl_rt_error(tcb, data, FL_IO_ERROR);
        it.it_interval.tv_sec = 0;
        it.it_interval.tv_nsec = 0;
        it.it_value.tv_sec = INT(data) / 1000;
        it.it_value.tv_nsec = (INT(data) % 1000) * 1000000;
        if(timerfd_settime(fd, 0, &it, NULL) < 0) 
            fl_rt_error(tcb, fl_nil, FL_IO_ERROR);
        break; }
    case FL_CHILD: { /* data: pid */
        struct child *c = (struct child *)malloc(sizeof(struct child));
        c->running = 1;
        c->tcb = tcb;
        c->pid = INT(data);
        c->var = var;
        pthread_mutex_lock(&child_mutex); /* XXX use atomic */
        c->next = children;
        children = c;
        pthread_mutex_unlock(&child_mutex);
        fd = 0;
        break; }
    }
    if(fd == - 1) {
        tcb->listening = CDR(n);
        CDR(n) = fl_nil;
        unref(tcb, n); 
        fl_rt_error(tcb, data, FL_IO_ERROR);
        return 0;
    }
    CAR(n1) = MKINT(fd);
    if(eid != NULL) fl_assign(tcb, MKINT(fd), eid);
    return update_events(tcb, 0);
}

void fl_cancel_timer(FL_TCB *tcb, FL_VAL id)
{
    FL_VAL *prev = &tcb->listening;
    FL_VAL ll = tcb->listening;
    while(ll != fl_nil) {
        if(CAR(CAR(ll)) == id) {
            *prev = CDR(ll);
            CDR(ll) = fl_nil;
            FL_VAL var = CAR(CDR(CAR(ll)));
            if(CAR(var) == var) fl_assign(tcb, fl_false, var);
            close(INT(CAR(CAR(ll))));
            unref(tcb, ll);
            return;
        } 
        prev = &CDR(ll);
        ll = CDR(ll);
    }
}
